import casadi as ca
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import linprog


# for people who cannot see the plots
if mpl.get_backend() == 'agg':
    mpl.use('WebAgg')
print(f"backend: {mpl.get_backend()}")


np.random.seed(5)             # seed for reproducible "randomness"
## 4b)
N = 30
x = np.linspace(0,5,N)
y = 3*x + 4                   # true output
y = y + np.random.randn(N)    # add noise

# visualize
plt.figure(1)
plt.plot(x, y, 'ko', label=r'measurements no outliers')

## 4c)
J = np.column_stack((x, np.ones(N)))
ab = np.linalg.solve(J.T @ J, J.T @ y)  # solve linear system
plt.plot(x, ab[0]*x + ab[1], 'b', label=r'L2 norm (no outliers)')

## 4d)
# introduce outliers
y2      = np.copy(y)
y2[0]   = 25
y2[4]   = 30
y2[-1]  = 10
plt.plot(x, y2, 'kx', label=r'measurements with outliers')

ab2 = np.linalg.solve(J.T @ J, J.T @ y2)  # solve
plt.plot(x, ab2[0]*x + ab2[1], 'r', label=r'L2 norm (outliers)')

## TASK 5
# 5b)

f = np.concatenate(([0, 0], np.ones(N)))

A = np.block([
    [ J, -np.eye(N)],
    [-J, -np.eye(N)]
])

b  = np.concatenate((y , -y))   # without outliers
b2 = np.concatenate((y2, -y2))  # with outliers

ab3 = linprog( f, A, b ).x
ab4 = linprog( f, A, b2).x

plt.plot(x, ab3[0]*x + ab3[1], 'm', label=r'L1 norm (no outliers)')
plt.plot(x, ab4[0]*x + ab4[1], 'c', label=r'L1 norm (outliers)')

## 5c)

opti = ca.Opti()
ab_ca = opti.variable(2)    # parameters
s = opti.variable(N)        # slacks

opti.minimize(ca.sum1(s))
opti.subject_to(-s <= J@ab_ca - y)
opti.subject_to(J@ab_ca - y <= s)
opti.solver('ipopt')
sol = opti.solve()
ab_ca_sol = sol.value(ab_ca)

plt.plot(x, ab_ca_sol[0]*x + ab_ca_sol[1], 'y--', label=r'casadi L1 (no outliers)')

## show plots
plt.legend(loc='best')
plt.show()
